@@ -21,6 +21,12 @@
 
 #include "llvm/Config/llvm-config.h"
 
+#if LLVM_VERSION_MAJOR >= 17
+#  include "llvm/TargetParser/Host.h"
+#else
+#  include "llvm/Support/Host.h"
+#endif
+
 #include "clang/AST/ASTConsumer.h"
 #include "clang/AST/Decl.h"
 #include "clang/AST/DeclCXX.h"
@@ -42,7 +48,6 @@
 #include "llvm/ADT/STLExtras.h"
 #include "llvm/Option/ArgList.h"
 #include "llvm/Support/FileSystem.h"
-#include "llvm/Support/Host.h"
 #include "llvm/Support/Path.h"
 #include "llvm/Support/raw_ostream.h"
 
@@ -56,7 +61,7 @@
 #  define CASTXML_OWNS_OSTREAM
 #endif
 
-#if LLVM_VERSION_MAJOR > 9
+#if LLVM_VERSION_MAJOR >= 10
 #  define CASTXML_MAKE_UNIQUE std::make_unique
 #else
 #  define CASTXML_MAKE_UNIQUE llvm::make_unique
@@ -602,7 +607,12 @@ runClangCreateDiagnostics(const char* const* argBeg, const char* const* argEnd)
 #if LLVM_VERSION_MAJOR > 3 ||                                                 \
   LLVM_VERSION_MAJOR == 3 && LLVM_VERSION_MINOR >= 7
   llvm::opt::InputArgList args(opts->ParseArgs(
-    llvm::makeArrayRef(argBeg, argEnd), missingArgIndex, missingArgCount));
+#  if LLVM_VERSION_MAJOR >= 16
+    llvm::ArrayRef(argBeg, argEnd),
+#  else
+    llvm::makeArrayRef(argBeg, argEnd),
+#  endif
+    missingArgIndex, missingArgCount));
   clang::ParseDiagnosticArgs(*diagOpts, args);
 #else
   std::unique_ptr<llvm::opt::InputArgList> args(
@@ -675,7 +685,9 @@ static int runClangImpl(const char* const* argBeg, const char* const* argEnd,
       const char* const* cmdArgEnd = cmdArgBeg + cmd->getArguments().size();
       if (clang::CompilerInvocation::CreateFromArgs(
             CI->getInvocation(),
-#if LLVM_VERSION_MAJOR > 9
+#if LLVM_VERSION_MAJOR >= 16
+            llvm::ArrayRef(cmdArgBeg, cmdArgEnd),
+#elif LLVM_VERSION_MAJOR >= 10
             llvm::makeArrayRef(cmdArgBeg, cmdArgEnd),
 #else
             cmdArgBeg, cmdArgEnd,
@@ -774,7 +786,13 @@ int runClang(const char* const* argBeg, const char* const* argEnd,
                   msvc_lang = std::strtol(msvc_lang_str.c_str(), nullptr, 10);
                 }
               }
-              if (msvc_lang >= 201703L) {
+              if (msvc_lang >= 202002L) {
+#if LLVM_VERSION_MAJOR >= 11
+                args.push_back("-std=c++20");
+#else
+                args.push_back("-std=c++17");
+#endif
+              } else if (msvc_lang >= 201703L) {
                 args.push_back("-std=c++17");
               } else {
                 args.push_back("-std=c++14");
@@ -815,7 +833,15 @@ int runClang(const char* const* argBeg, const char* const* argEnd,
             std_date = 0;
           }
           std_flag += "++";
-          if (std_date >= 201703L) {
+          if (std_date >= 202002L) {
+#if LLVM_VERSION_MAJOR >= 11
+            std_flag += "20";
+#elif LLVM_VERSION_MAJOR >= 5
+            std_flag += "17";
+#else
+            std_flag += "1z";
+#endif
+          } else if (std_date >= 201703L) {
 #if LLVM_VERSION_MAJOR >= 5
             std_flag += "17";
 #else
